pragma solidity >=0.8.7 <0.9.0;

contract TurtlesNFT {

    event Transfer(address indexed _from, address indexed _to, uint256 indexed _tokenId);
    event Approval(address indexed _owner, address indexed _approved, uint256 indexed _tokenId);
    event ApprovalForAll(address indexed _owner, address indexed _operator, bool _approved);

    uint256 randNonce;

    uint256 numTokens;
    uint256 remaining;
    mapping (uint256 => address) private tokenToOwner;
    mapping (address => uint256) private ownedTokens;
    mapping (uint256 => string) private tokenLinks;

    struct approval {
        address approved;
        uint256 tokenId;
    }

    mapping (address => approval) approvals;

    constructor() {
        tokenLinks[0] = "QmZCwuN2ErndLfQ4mvqzTDE88FTSMNrTRw7kbx7moZjkG3";
        tokenLinks[1] = "QmSwgxdf91Wae9x2EaQnpQAbcEoE8Z8QaPLbXoKJwcVeLu";
        tokenLinks[2] = "QmTM8X1iqtifLMjtSAfWuMKovnWNj5H2jyupTLns16Z6Et";
        tokenLinks[3] = "QmQN1FuDDif9swMc7eGEKfww6a1LQSCW5JorLvnoaYhnxG";
        tokenLinks[4] = "QmSzEjPWqHcuiF2wdKy4mApvpAJYPLCxQVhd8BLn7vowtJ";
        tokenLinks[5] = "QmZsGZ8evF3weMHPHGS6vBA4SMewUEd1LQjBGXiY9R3dY6";
        numTokens = 6;
        remaining = 6;
        randNonce = 420;
    }

    function balanceOf(address _owner) external view returns (uint256) {
        return ownedTokens[_owner];
    }

    function ownerOf(uint256 _tokenId) external view returns (address) {
        return tokenToOwner[_tokenId];
    }

    modifier onlyOwnerOf(uint256 _tokenId) {
        require(msg.sender == tokenToOwner[_tokenId]);
        _;
    }

    function transferFrom(address _from, address _to, uint256 _tokenId) external payable {
        require(_from == msg.sender || (approvals[_from].approved == msg.sender && approvals[_from].tokenId == _tokenId));
        require(_from == tokenToOwner[_tokenId]);
        tokenToOwner[_tokenId] = _to;
        ownedTokens[_from]--;
        ownedTokens[_to]++;
        emit Transfer(_from, _to, _tokenId);
    }

    function approve(address _approved, uint256 _tokenId) external payable onlyOwnerOf(_tokenId) {
        approvals[msg.sender] = approval(_approved, _tokenId);
        emit Approval(msg.sender, _approved, _tokenId);
    }

    function getApproved(uint256 _tokenId) external view returns (address) {
        return approvals[tokenToOwner[_tokenId]].approved;
    }

    // also in standard:
    // safeTransferFrom
    // setApprovalForAll
    // isApprovedForAll

    // not part of standard

    function transfer(address _to, uint256 _tokenId) external onlyOwnerOf(_tokenId) {
        tokenToOwner[_tokenId] = _to;
        ownedTokens[msg.sender]--;
        ownedTokens[_to]++;
        emit Transfer(msg.sender, _to, _tokenId);
    }

    function random(uint256 _modulus) internal returns (uint256) {
        return uint(keccak256(abi.encodePacked(block.timestamp, msg.sender, randNonce++))) % _modulus;
    }

    function mint() external returns (uint256) {
        require(remaining >= 0);
        uint256[] memory available = new uint[](remaining);
        uint256 counter = 0;
        for (uint i = 0; i < numTokens; i++) {
            if (tokenToOwner[i] == address(0x0)) {
                available[counter++] = i;
            }
        }
        assert(counter == remaining);
        uint256 tokenId = available[random(remaining)];
        tokenToOwner[tokenId] = msg.sender;
        ownedTokens[msg.sender]++;
        remaining--;
        return tokenId;
    }

    function getLink(uint256 _tokenId) public view returns (string memory) {
        require(_tokenId < numTokens);
        return tokenLinks[_tokenId];
    }

}

