pragma solidity >=0.8.7 <0.9.0;

contract TurtlesToken {

    event Transfer(address indexed _from, address indexed _to, uint256 _value);
    event Approval(address indexed _owner, address indexed _spender, uint256 _value);

    uint256 totalTokens;
    uint256 remaining;

    struct approval {
        address spender;
        uint256 value;
    }

    mapping (address => uint256) addrToTokens;
    mapping (address => approval) approvals;

    address owner;
    uint256 price;

    constructor(uint256 _totalTokens, uint256 _price) {
        owner = msg.sender;
        totalTokens = _totalTokens;
        remaining = _totalTokens;
        price = _price;
    }

    function name() public view returns (string memory) {
        return "TurtlesToken";
    }

    function symbol() public view returns (string memory) {
        return "TRT";
    }

    function decimals() public view returns (uint8) {
        return 8;
    }

    function totalSupply() public view returns (uint256) {
        return totalTokens;
    }

    function balanceOf(address _owner) public view returns (uint256 balance) {
        return addrToTokens[_owner];
    }

    function transfer(address _to, uint256 _value) public returns (bool success) {
        require(addrToTokens[msg.sender] >= _value);
        addrToTokens[_to] += _value;
        addrToTokens[msg.sender] -= _value;
        emit Transfer(msg.sender, _to, _value);
        return true;
    }

    function transferFrom(address _from, address _to, uint256 _value) public returns (bool success) {
        require(_from == msg.sender || (approvals[_from].spender == msg.sender && approvals[_from].value <= _value));
        addrToTokens[_to] += _value;
        addrToTokens[_from] -= _value;
        if (approvals[_from].spender == msg.sender) {
            approvals[_from].value -= _value;
        }
        emit Transfer(_from, _to, _value);
        return true;
    }

    function approve(address _spender, uint256 _value) public returns (bool success) {
        approvals[msg.sender] = approval(_spender, _value);
        emit Approval(msg.sender, _spender, _value);
        return true;
    }

    function allowance(address _owner, address _spender) public view returns (uint256 remaining) {
        require(approvals[_owner].spender == _spender);
        return approvals[_owner].value;
    }

    // these are not part of the ERC-20 standard

    modifier onlyOwner() {
        require(msg.sender == owner);
        _;
    }

    function getPrice() public view returns (uint256) {
        return price;
    }

    function setPrice(uint256 _price) public onlyOwner {
        price = _price;
    }

    function collect() public onlyOwner {
        (bool sent,) = msg.sender.call{ value: address(this).balance }("");
        require(sent, "Failed to collect");
    }

    function buy() public payable returns (uint256 bought) {
        require(msg.value >= price);
        uint256 amount = msg.value / price;
        require(amount <= remaining);
        remaining -= amount;
        addrToTokens[msg.sender] += amount;
        return amount;
    }

}

